; hacker port firmware
; version 0.3.1 date 2004-05-26
; part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
; Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

	TITLE		"hackerport firmware"

	include		"P18F442.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================

	extern		USB_R1_CNT, USB_R1_BUF
	extern		USB_R2_CNT, USB_R2_BUF
	extern		USB_R3_CNT, USB_R3_BUF
	extern		USB_Sx_CNT, USB_Sx_BUF



;===========================================================================================
; imported procedures
;===========================================================================================

	extern		DEBUG_INIT
	extern		USB_INIT, USB_TASK
	extern		MSG_INIT, MSG_PROC



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; storage for global bits (see Bits.inc)
BITS0		res	1

; current state of latched port (see IoPins.inc)
LATCHED_PORT	res	1

; general purpose registers
A1_REG		res	1
B1_REG		res	1
C1_REG		res	1
D1_REG		res	1
A2_REG		res	2
B2_REG		res	2
C2_REG		res	2
D2_REG		res	2
A4_REG		res	4

; sequence numbers for USB communication
SEQ_NO_RECV	res	1
SEQ_NO_SEND	res	1

; status of packet being processed
PROC_EP		res	1		; number of endpoint of which a packet is processed (0 if none)
PROC_PTR	res	2		; address of next byte to process
PROC_LEN	res	1		; length of data to process

; OLD: buffer for output port values and PWM values
OUT_PWM_BUF	res	0
OUT_A_VAL	res	1		; values for output ports
OUT_B_VAL	res	1
PWM_A_VAL	res	2		; values for PWM outputs (big endian)
PWM_B_VAL	res	2

; OLD: buffer for input port values and analog input values
IN_ANA_BUF	res	0
IN_A_VAL	res	1		; values from input ports
IN_B_VAL	res	1
ANA_A_VAL	res	2		; values from analog input (big endian)
ANA_B_VAL	res	2
ANA_C_VAL	res	2
ANA_D_VAL	res	2
ANA_E_VAL	res	2



;===========================================================================================
ram_main	UDATA			; normal RAM for main
;===========================================================================================

; backup registers for low priority interrupts
INT_WREG	res	1
INT_STATUS	res	1
INT_BSR		res	1



;===========================================================================================
entry		CODE			; code entry point
;===========================================================================================
ENTRY:
	nop				; used by ICD
	nop
	goto	INIT

;===========================================================================================
int_hi		CODE			; high priority interrupts
;===========================================================================================
HIGH_INTERRUPT:
	reset				; something went completely wrong, so start over

;===========================================================================================
int_lo		CODE			; low priority interrupts
;===========================================================================================
LOW_INTERRUPT:
	reset				; something went completely wrong, so start over



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; initialization
;-------------------------------------------------------------------------------------------
INIT:
; set default values for output pins
	bcf	PIN_LATCH_LE		; clear latched port latch enable
	bcf	PIN_OUT_A_LE		; clear OUTPUT A latch enable
	bcf	PIN_OUT_B_LE		; clear OUTPUT B latch enable
	bsf	PIN_IN_A_nOE		; clear INPUT A output enable
	bsf	PIN_IN_B_nOE		; clear INPUT B output enable
	clrf	PORT_BUS_O		; set data bus to 0x00
	bcf	PIN_PIC_PGD_O		; clear PICPROG PGD
	bsf	PIN_USB_nRST		; clear USBN9604 reset
	bcf	PIN_USB_A0		; clear USBN9604 address bit
	bsf	PIN_USB_nRD		; set USBN9604 read
	bsf	PIN_USB_nWR		; set USBN9604 write
	bsf	PIN_USB_nCS		; set USBN9604 chis select
; set up analog inputs
	movlw	b'10000001'		; FOSC/64 = 1.6us at 40MHz, select AN0
	movwf	ADCON0			; no conversion active, analog input module powered up
	movlw	b'01000010'		; left justified values, AN0..AN4 (RA0..3+RA5) active
	movwf	ADCON1
; define inputs/outputs
	setf	TRISA			; RA0..6 input (RA7 does not exist)
	clrf	TRISB			; RB0..5 output (RB6..7 for ICD2 connection)
	movlw	b'10000000'		; RC0+RC5 output, RC1..2 (PWM) output
	movwf	TRISC			; RC3+RC4 (SCL+SDA) output, RC6 (TX) output, RC7 (RX) input
	setf	TRISD			; RD0..7 input
	clrf	TRISE			; RE0..2 output (RE3..7 do not exist), parallel slave port off
; initialize timers
	movlw	b'01111111'		; timer 0 off
	movwf	T0CON
	movlw	b'00000000'		; timer 1 off
	movwf	T1CON
	movlw	b'00000110'		; timer 2 on, prescaler 1:16, postscaler 1:1
	movwf	T2CON
	movlw	249			; timer 2: 16*250=4000 cycles, period = 2.5 kHz at 40 MHz
	movwf	PR2
	movlw	b'00000000'		; timer 3 off
	movwf	T3CON
; set up PWM outputs
	clrf	CCPR1L			; CCP1: duty cycle 0
	movlw	b'00001100'		; CCP1: PWM mode
	movwf	CCP1CON
	clrf	CCPR2L			; CCP2: duty cycle 0
	movlw	b'00001100'		; CCP2: PWM mode
	movwf	CCP2CON
; set up I2C in master mode
	movlw	b'10000000'		; no slew rate control, SMBus off, clear status flags
	movwf	SSPSTAT
	movlw	b'00101000'		; no collision yet, enable MSSP, master mode
	movwf	SSPCON1
	movlw	b'00000000'		; no acknowledge yet, recv off, stop off, repeated start off, start off
	movwf	SSPCON2
	movlw	0x3F			; 100kHz
	movwf	SSPADD
; set up USART
	movlw	b'00000110'		; 8 bit, async mode, high speed, shift reg empty
	movwf	TXSTA
	movlw	b'10010000'		; port enabled, 8 bit, recv enabled, no addr detect
	movwf	RCSTA
	movlw	22			; 115200 kbps (1,4% error)
	movwf	SPBRG
; initialize latched port
	bcf	PINL_JTAG_TRST		; JTAG bits to 0
	bcf	PINL_JTAG_TCLK
	bcf	PINL_JTAG_TMS
	bcf	PINL_JTAG_TDI
	bcf	PINL_PIC_VPP		; PICPROG bits to 0
	bcf	PINL_PIC_MCLR
	bcf	PINL_PIC_PGC
	bcf	PINL_LED_13V		; USB LED / 13V generator clock off
	call	LATCHED_PORT_OUT
; initialize output ports
	clrf	PORT_BUS_O		; output 0x00 on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
; initialize PWM outputs
	clrf	CCPR1L			; PWM A to 0% duty cycle
	bcf	CCP1CON,DC1B1
	bcf	CCP1CON,DC1B0
	clrf	CCPR2L			; PWM B to 0% duty cycle
	bcf	CCP2CON,DC2B1
	bcf	CCP2CON,DC2B0
; initialize global variables
	clrf	BITS0
	clrf	SEQ_NO_RECV		; clear sequence numbers
	clrf	SEQ_NO_SEND
; call initialization procedures
	call	DEBUG_INIT
	call	USB_INIT
	call	MSG_INIT
; initialize local variables
	setf	ANA_A_VAL+0		; impossible value (last 6 bits are always 0)
	setf	ANA_A_VAL+1		; - this leads to a change detection after first read
	clrf	PROC_EP			; not processing a packet
; initialize interrupts
;!!!
; reset watchdog
	clrwdt
; initialization done, start program
	goto	MAIN

;-------------------------------------------------------------------------------------------
; output value of LATCHED_PORT to latched port
; input:	LATCHED_PORT		new state for the latched port
; output:	-
; changes:	-
;-------------------------------------------------------------------------------------------
LATCHED_PORT_OUT:
	movff	LATCHED_PORT,PORT_BUS_O	; output data
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_LATCH_LE		; set latched port latch enable
	bcf	PIN_LATCH_LE		; clear latched port latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	return

;-------------------------------------------------------------------------------------------
; main program
;-------------------------------------------------------------------------------------------
MAIN:

; wait 1 ms
	movlw	10
	movwf	B1_REG
MAIN_WAIT:
	clrwdt
	movlw	200
	movwf	A1_REG
MAIN_WAIT_I:
	nop
	nop
	decfsz	A1_REG
	bra	MAIN_WAIT_I
	decfsz	B1_REG
	bra	MAIN_WAIT

; main loop
MAIN_LOOP:

; call USB
	call	USB_TASK

; check if not processing a packet
	movf	PROC_EP			; jump if still processing a packet
	bnz	MAIN_R_END
; check if received the next packet on EP1
	movf	USB_R1_CNT		; jump if nothing received on EP1
	bz	MAIN_R1_END
	movff	USB_R1_BUF,WREG		; jump if wrong sequence number
	cpfseq	SEQ_NO_RECV
	bra	MAIN_R1_SEQ		; (packet's seq. no. in WREG)
; next packet was received on EP1
	incf	SEQ_NO_RECV		; increment sequence number
	movlw	1			; processing packet of EP1
	movwf	PROC_EP
	movlw	high (USB_R1_BUF + 1)	; remember address (behind sequence number)
	movwf	PROC_PTR+1
	movlw	low (USB_R1_BUF + 1)
	movwf	PROC_PTR+0
	decf	USB_R1_CNT,W		; remember length (without sequence number)
	movwf	PROC_LEN
	bra	MAIN_R_END
; skip packet if sequence number of packet is "too small"
MAIN_R1_SEQ:
	subwf	SEQ_NO_RECV,W		; sequence number of packet is still in WREG
	movwf	A1_REG			; A1_REG = expected seq. no. - packet's seq. no.
	btfss	A1_REG,7		; if A1_REG > 0 (i.e. packet's seq. no. "too small"), ...
	clrf	USB_R1_CNT		; ..., skip this packet
MAIN_R1_END:
; check if received the next packet on EP2
	movf	USB_R2_CNT		; jump if nothing received on EP2
	bz	MAIN_R2_END
	movff	USB_R2_BUF,WREG		; jump if wrong sequence number
	cpfseq	SEQ_NO_RECV
	bra	MAIN_R2_SEQ		; (packet's seq. no. in WREG)
; next packet was received on EP2
	incf	SEQ_NO_RECV		; increment sequence number
	movlw	2			; processing packet of EP2
	movwf	PROC_EP
	movlw	high (USB_R2_BUF + 1)	; remember address (behind sequence number)
	movwf	PROC_PTR+1
	movlw	low (USB_R2_BUF + 1)
	movwf	PROC_PTR+0
	decf	USB_R2_CNT,W		; remember length (without sequence number)
	movwf	PROC_LEN
	bra	MAIN_R_END
; skip packet if sequence number of packet is "too small"
MAIN_R2_SEQ:
	subwf	SEQ_NO_RECV,W		; sequence number of packet is still in WREG
	movwf	A1_REG			; A1_REG = expected seq. no. - packet's seq. no.
	btfss	A1_REG,7		; if A1_REG > 0 (i.e. packet's seq. no. "too small"), ...
	clrf	USB_R2_CNT		; ..., skip this packet
MAIN_R2_END:
; check if received the next packet on EP3
	movf	USB_R3_CNT		; jump if nothing received on EP3
	bz	MAIN_R3_END
	movff	USB_R3_BUF,WREG		; jump if wrong sequence number
	cpfseq	SEQ_NO_RECV
	bra	MAIN_R3_SEQ		; (packet's seq. no. in WREG)
; next packet was received on EP3
	incf	SEQ_NO_RECV		; increment sequence number
	movlw	3			; processing packet of EP3
	movwf	PROC_EP
	movlw	high (USB_R3_BUF + 1)	; remember address (behind sequence number)
	movwf	PROC_PTR+1
	movlw	low (USB_R3_BUF + 1)
	movwf	PROC_PTR+0
	decf	USB_R3_CNT,W		; remember length (without sequence number)
	movwf	PROC_LEN
	bra	MAIN_R_END
; skip packet if sequence number of packet is "too small"
MAIN_R3_SEQ:
	subwf	SEQ_NO_RECV,W		; sequence number of packet is still in WREG
	movwf	A1_REG			; A1_REG = expected seq. no. - packet's seq. no.
	btfss	A1_REG,7		; if A1_REG > 0 (i.e. packet's seq. no. "too small"), ...
	clrf	USB_R3_CNT		; ..., skip this packet
MAIN_R3_END:
; if we get here, no packet matches the expected sequence number
; check if received a packet on all 3 EPs
	movf	USB_R1_CNT
	bz	MAIN_R_END
	movf	USB_R2_CNT
	bz	MAIN_R_END
	movf	USB_R3_CNT
	bz	MAIN_R_END
; if we get here, there are 3 packets with a "too high" sequence number
;  - e.g. we missed a packet
; use "smallest" packet's sequence number as new one to synchronize
MAIN_R_SYNC:
	incf	SEQ_NO_RECV		; increment sequence number until it matches any packet
	movff	USB_R1_BUF,WREG		; match on EP1
	xorwf	SEQ_NO_RECV,W
	bz	MAIN_R_END
	movff	USB_R2_BUF,WREG		; match on EP2
	xorwf	SEQ_NO_RECV,W
	bz	MAIN_R_END
	movff	USB_R3_BUF,WREG		; match on EP3
	xorwf	SEQ_NO_RECV,W
	bz	MAIN_R_END
	bra	MAIN_R_SYNC
MAIN_R_END:

; check if a packet to process is availabale
	movf	PROC_EP			; jump if no packet is being processed
	bz	MAIN_PROC_END
; process packet
	call	MSG_PROC
; check if packet done
	movf	PROC_LEN		; jump if packet not done yet
	bnz	MAIN_PROC_END
; cleanup
	movlw	1			; if packet was received on EP1, ...
	cpfseq	PROC_EP
	bra	MAIN_PROC1_END
	clrf	USB_R1_CNT		; ..., clear receive buffer of EP1
MAIN_PROC1_END:
	movlw	2			; if packet was received on EP2, ...
	cpfseq	PROC_EP
	bra	MAIN_PROC2_END
	clrf	USB_R2_CNT		; ..., clear receive buffer of EP2
MAIN_PROC2_END:
	movlw	3			; if packet was received on EP3, ...
	cpfseq	PROC_EP
	bra	MAIN_PROC3_END
	clrf	USB_R3_CNT		; ..., clear receive buffer of EP3
MAIN_PROC3_END:
	clrf	PROC_EP			; no packet is being processed
MAIN_PROC_END:

; if we are allowed to draw power from USB, turn on LED and run 13V generator
	btfss	BIT_USB_PWR_OK
	bra	MAIN_LED_OFF		; jump if we may not draw power from USB
	btfsc	PINL_LED_13V
	bra	MAIN_LED_OFF		; jump if LED is on (13V generator needs toggling bit)
	bsf	PINL_LED_13V		; turn on LED
	bra	MAIN_LED_OUT
MAIN_LED_OFF:
	bcf	PINL_LED_13V		; turn off LED
MAIN_LED_OUT:
	call	LATCHED_PORT_OUT	; update state of latched port

; reset watchdog
	clrwdt
; go on with main loop
	bra	MAIN_LOOP



;===========================================================================================
; global variables
;===========================================================================================
	global		BITS0
	global		LATCHED_PORT
	global		A1_REG, B1_REG, C1_REG, D1_REG
	global		A2_REG, B2_REG, C2_REG, D2_REG
	global		A4_REG
	global		SEQ_NO_RECV, SEQ_NO_SEND
	global		PROC_PTR, PROC_LEN



;===========================================================================================
; global procedures
;===========================================================================================

	global		LATCHED_PORT_OUT



	end
